#include "scene.h"

#include "../shaders/shader.h"
#include "../shaders/intersectionshader.h"
#include "../shaders/depthshader.h"
#include "../shaders/pathshader.h"
#include "../shaders/directshader.h"
#include "../shaders/normalshader.h"
#include "../shaders/irradiance.h"
#include "../../build/Irradiance_first.h"

#include "../cameras/ortographic.h"
#include "../cameras/perspective.h"

#include "../materials/phong.h"
#include "../materials/mirror.h"
#include "../materials/transmissive.h"

#include "../shapes/sphere.h"
#include "../shapes/triangle.h"

#include <fstream>
#include <unordered_map>

Camera & Scene::getCamera() { 
	return *cam; 
}

bool Scene::loadScene(const char* filename, Shader *&shader_out)
{
	lightSourceList = new std::vector<LightSource>;
	objectsList = new std::vector<Shape*>;
	areaLSList = new std::vector<Shape*>;
	string line;

	std::ifstream inFile;
	inFile.open(filename);
	if (!inFile) {
		std::cerr << "Unable to open file " << filename << std::endl;
		return false;
	}

	bool data_for_film = false;
	size_t width, height;

	std::unordered_map<std::string, Material*> materials;

	while (std::getline(inFile, line)) {

		// Exclude comments and new lines
		if (line.compare(0, 2, "//") && !line.empty()){
			
			// Get first word specifying the entity
			std::stringstream iss(line);
			string entity;
			iss >> entity;

			// If it's a Film
			if (entity.compare("Film") == 0) {
				iss >> width >> height;
				

				std::cout << entity << " " << width << " " << height << std::endl;

			}
			// It's a shader
			else if (entity.compare("Shader") == 0) {
				string shaderType;
				iss >> shaderType;

				// Case path shader
				if (shaderType.compare("Path") == 0) {
					double r, g, b;
					size_t nSpp, nSamplesArea, max_path_depth;
					bool indirectOnly;
					iss >> r >> g >> b >> nSpp >> nSamplesArea >> max_path_depth >> std::boolalpha >> indirectOnly;
					Vector3D bgColor(r, g, b);
					shader_out = new PathShader(bgColor, nSpp, nSamplesArea, max_path_depth, indirectOnly);
					std::cout << entity << " " << shaderType << " " << bgColor << " " << max_path_depth << std::endl;
					data_for_film = true;
				}
				else if (shaderType.compare("Direct") == 0) {
					double r, g, b;
					size_t nSpp, nSamplesHemi, nSamplesArea;
					iss >> r >> g >> b >> nSpp >> nSamplesHemi >> nSamplesArea;
					Vector3D bgColor(r, g, b);
					shader_out = new DirectShader(bgColor, nSpp, nSamplesHemi, nSamplesArea);
					std::cout << entity << " " << shaderType << " " << bgColor << std::endl;
					data_for_film = true;
				}
				else if (shaderType.compare("Irradiance") == 0) {
					double r, g, b;
					size_t nSamplesArea, max_path_depth, nSamplesHemi;
					bool points;
					bool indirectOnly;
					iss >> r >> g >> b >> nSamplesArea  >> max_path_depth >> points >> std::boolalpha >> indirectOnly >> nSamplesHemi;
					Vector3D bgColor(r, g, b);
					std::cout << points << std::endl;
					shader_out = new irradiance(bgColor, max_path_depth, points, nSamplesArea, indirectOnly, nSamplesHemi);
					std::cout << entity << " " << shaderType << " " << bgColor << " " << max_path_depth << std::endl;
					std::cout << "Irradiance 100%" << std::endl;
					data_for_film = true;
				}
				else if (shaderType.compare("Depth") == 0) {
					double c_r, c_g, c_b, max_dist, bkg_r, bkg_g, bkg_b;
					bool color_map;
					iss >> c_r >> c_g >> c_b >> max_dist >> bkg_r >> bkg_g >> bkg_b >> std::boolalpha >> color_map;
					Vector3D depthColor(c_r, c_g, c_b);
					Vector3D bgColor(bkg_r, bkg_g, bkg_b);
					shader_out = new DepthShader(depthColor, max_dist, bgColor, color_map);
					std::cout << entity << " " << shaderType << " " << depthColor << " " << max_dist << " " << bgColor << " " << color_map << std::endl;
					data_for_film = true;
				}
				else if (shaderType.compare("Intersection") == 0) {
					double c_r, c_g, c_b, bkg_r, bkg_g, bkg_b;
					iss >> c_r >> c_g >> c_b >> bkg_r >> bkg_g >> bkg_b;
					Vector3D itsColor(c_r, c_g, c_b);
					Vector3D bgColor(bkg_r, bkg_g, bkg_b);
					shader_out = new IntersectionShader(itsColor, bgColor);
					std::cout << entity << " " << shaderType << " " << itsColor << " " << bgColor << std::endl;
					data_for_film = true;
				}
				else if (shaderType.compare("Normal") == 0) {
					double bkg_r, bkg_g, bkg_b;
					iss >> bkg_r >> bkg_g >> bkg_b;
					Vector3D bgColor(bkg_r, bkg_g, bkg_b);
					shader_out = new NormalShader(bgColor);
					std::cout << entity << " " << shaderType << " " << bgColor << std::endl;
					data_for_film = true;
				}
				else {
					std::cout << "Unknown shader type " << shaderType << std::endl;
				}

			}
			// It's a camera
			else if (entity.compare("Camera") == 0) {
				string cameraType;
				iss >> cameraType;
				Matrix4x4 cameraToWorld; // By default, this gives an ID transform

				// Case perspective camera
				if (cameraType.compare("Perspective") == 0) {
					double fovDegrees;
					iss >> fovDegrees;
					double fovRadians = Utils::degreesToRadians(fovDegrees);
					
					if (data_for_film) {
						Film *film = new Film(width, height);
						cam = new PerspectiveCamera(cameraToWorld, fovRadians, *film);
					}
					else {
						std::cout << "Cannot initialize camera: Film not declared previously in the scene file" << std::endl;
					}
					std::cout << entity << " " << cameraType << " fov = " << fovDegrees << std::endl;
				}
				else {
					std::cout << "Unknown camera type " << cameraType << std::endl;
				}

			}
			// It's a material
			else if (entity.compare("Material") == 0) {
				string materialType;
				iss >> materialType;

				// Case phong material
				if (materialType.compare("Phong") == 0) {
					string name;
					iss >> name;
					double kd_r, kd_g, kd_b;
					double ks_r, ks_g, ks_b;
					double shininess;

					string next;
					iss >> next;

					Material *mat;
					if (next[0] == '.') {
						iss >> shininess;
						mat = new Phong(next, shininess);
					}
					else {
						kd_r = stod(next);
						iss >> kd_g >> kd_b >> ks_r >> ks_g >> ks_b >> shininess;
						mat = new Phong(Vector3D(kd_r, kd_g, kd_b), Vector3D(ks_r, ks_g, ks_b), shininess);
					}
					
					materials.insert(std::pair<string, Material*>(name, mat));
					std::cout << "Stored Phong " << name << " in the unnordered map" << std::endl;
				}
				// Case mirror material
				else if (materialType.compare("Mirror") == 0) {
					string name;
					iss >> name;

					Material *mat = new Mirror();
					materials.insert(std::pair<string, Material*>(name, mat));
					std::cout << "Stored Mirror " << name << " in the unnordered map" << std::endl;
				}
				// Case	transmissive material
				else if (materialType.compare("Transmissive") == 0) {
					string name;
					iss >> name;
					double ior, r, g, b;

					iss >> ior >> r >> g >> b;
					Material *mat = new Transmissive(ior, Vector3D(r, g, b));
					materials.insert(std::pair<string, Material*>(name, mat));
					std::cout << "Stored Transmissive " << name << " in the unnordered map" << std::endl;
				}
				else {
					std::cout << "Unknown material type " << materialType << std::endl;
				}

			}
			else if (entity.compare("Object") == 0) {

				string objectType;
				iss >> objectType;

				// Case sphere
				if (objectType.compare("Sphere") == 0) {
					double c_x, c_y, c_z, rad;
					string mat_name;
					bool isEmissive;

					iss >> c_x >> c_y >> c_z >> rad >> mat_name >> std::boolalpha >> isEmissive;
					Material *mat = materials.at(mat_name);

					// Define and place the sphere
					Matrix4x4 sphereTransform;
					sphereTransform = sphereTransform.translate(Vector3D(c_x, c_y, c_z));
					Shape *s = new Sphere(rad, sphereTransform, mat);
					if (isEmissive) {
						double e_r, e_g, e_b;
						iss >> e_r >> e_g >> e_b;
						s->setEmission(Vector3D(e_r, e_g, e_b));
						areaLSList->push_back(s);
					}
					// Store the objects in the object list
					objectsList->push_back(s);
					std::cout << entity << objectType << Vector3D(c_x, c_y, c_z) << rad << mat_name << std::endl;

				}
				// Case triangle
				else if (objectType.compare("Triangle") == 0) {
					double p0_x, p0_y, p0_z;
					double p1_x, p1_y, p1_z;
					double p2_x, p2_y, p2_z;
					string mat_name;
					bool isEmissive;

					iss >> p0_x >> p0_y >> p0_z >> p1_x >> p1_y >> p1_z >> p2_x >> p2_y >> p2_z >> mat_name >> std::boolalpha >> isEmissive;
					Material *mat = materials.at(mat_name);

					Vector3D p0(p0_x, p0_y, p0_z), p1(p1_x, p1_y, p1_z), p2(p2_x, p2_y, p2_z);
					Shape *triangle = new Triangle(p0, p1, p2, mat);
					if (isEmissive) {
						double e_r, e_g, e_b;
						iss >> e_r >> e_g >> e_b;
						triangle->setEmission(Vector3D(e_r, e_g, e_b));
						areaLSList->push_back(triangle);
					}
					objectsList->push_back(triangle);

					std::cout << entity << objectType << p0 << p1 << p2 << std::endl;
				}
				else {
					std::cout << "Unknown object type " << objectType << std::endl;
				}

			}
			else if (entity.compare("Pointlight") == 0) {

				double p_x, p_y, p_z, i_r, i_g, i_b;

				iss >> p_x >> p_y >> p_z >> i_r >> i_g >> i_b;

				LightSource light(Vector3D(p_x, p_y, p_z), Vector3D(i_r, i_g, i_b));
				lightSourceList->push_back(light);

				std::cout << entity << Vector3D(p_x, p_y, p_z) << Vector3D(i_r, i_g, i_b) << std::endl;
			}
			else {
				std::cout << "Unknown entity " << entity << std::endl;
			}

			
		}
	}

	inFile.close();
	return true;
}


bool Scene::hasIntersection(const Ray &cameraRay)
{
	// For each object on the scene...
	for (size_t objIndex = 0; objIndex < objectsList->size(); objIndex++)
	{
		// ..test the camera ray intersection with the current object
		const Shape *obj = objectsList->at(objIndex);
		if (obj->shadowRayIntersect(cameraRay))
			return true;
	}
	return false;
}

bool Scene::getClosestIntersection(const Ray &cameraRay, Intersection &its)
{
	// Variable to store the intersection detection
	bool intersection = false;

	// For each object on the scene...
	for (size_t objIndex = 0; objIndex < objectsList->size(); objIndex++)
	{
		// ..test the camera ray intersection with the current object
		const Shape *obj = objectsList->at(objIndex);
		if (obj->rayIntersect(cameraRay, its))
			intersection = true;
	}

return intersection;
}