#include "utils.h"

Utils::Utils()
{ }

double interpolate(double val, double y0, double x0, double y1, double x1 )
{
    return (val-x0)*(y1-y0)/(x1-x0) + y0;
}

double getRed(double value)
{
    if (value > 0.5)
        return interpolate( value, 0.0, 0.5, 1.0, 1.0 );
    else
        return 0.0;
}

double getGreen(double value)
{
    if (value < 0.25)
        return 0.0;
    else if (value < 0.5)
        return interpolate(value, 0.0, 0.25, 1, 0.5);
    else if (value < 0.75)
        return interpolate(value, 1, 0.5, 0, 0.75);
    else
        return 0;
}

double getBlue(double value)
{
    if (value < 0.5)
        return interpolate(value, 1.0, 0.0, 0.0, 0.5);
    else
        return 0;
}


Vector3D Utils::scalarToRGB(double scalar)
{
    return Vector3D( getRed(scalar),
                getGreen(scalar),
                getBlue(scalar) );
}

double Utils::degreesToRadians(double degrees)
{
    return degrees * M_PI / 180.0;
}

Vector3D Utils::multiplyPerCanal(const Vector3D &v1, const Vector3D &v2)
{
    return Vector3D(v1.x*v2.x, v1.y*v2.y, v1.z*v2.z);
}

Vector3D Utils::computeReflectionDirection(const Vector3D &rayDirection, const Vector3D &normal)
{
    // Compute the perfect reflection direction
    return normal * 2 * dot(normal, -rayDirection) + rayDirection;
}

bool Utils::isTotalInternalReflection(const double &eta, const double &cosThetaI,
                                      double &cosThetaT_out)
{
    // Compute the cos of the angle of the transmited ray with the normal
    // http://docslide.com.br/documents/luis-paulo-santos-computacao-grafica-0809-computacao-grafica-ray-tracing.html
    double cosThetaTSq = 1 - (1/(eta*eta)) * (1 - cosThetaI*cosThetaI );

    // If the squared cosinus is greater than zero, then we have transmission...
    if (cosThetaTSq > 0)
    {
        // ... and we can compute the cosinus angle of the transmission ray with -normal
        cosThetaT_out = sqrt(cosThetaTSq);
        return false;
    }
    // Otherwise, we are in "total internal reflection" (TIR) conditions, meaning that
    //  there is no transmission ...
    else
    {
        // ... and we return true (meaning that there is TIR)
        return true;
    }
}

Vector3D Utils::computeTransmissionDirection(const Ray &r, const Vector3D &normal,
                                             const double &eta, const double &cosThetaI,
                                             const double &cosThetaT)
{
    // Compute the transmission direction
    return r.d / eta - normal * (cosThetaT - cosThetaI / eta);
}

double Utils::computeReflectanceCoefficient(const double &eta, const double &cosThetaI,
                                            const double &cosThetaT)
{
    // 1 - Fresnel reflectance for parallel polarized light
    double reflectanceParllel = (cosThetaI - eta * cosThetaT) /
            (cosThetaI + eta * cosThetaT);

    // 2 - Fresnel reflectance for perpendicular polarized light
    double reflectancePerpendicular = (eta * cosThetaI - cosThetaT) /
            (eta * cosThetaI + cosThetaT);

    // 3 - The reflectance coefficient is then given by the average
    return 0.5f * ( reflectanceParllel * reflectanceParllel +
                    reflectancePerpendicular * reflectancePerpendicular );
}
