#include <iostream>
#include <stdlib.h> /* srand, rand */
#include <vector>

#include "core/film.h"
#include "core/matrix4x4.h"
#include "core/ray.h"
#include "core/utils.h"
#include "core/tester.h"
#include "core/scene.h"

#include "shapes/sphere.h"

#include "cameras/ortographic.h"
#include "cameras/perspective.h"

#include "materials/phong.h"
#include "materials/mirror.h"
#include "materials/transmissive.h"

#include "lightsources/lightsource.h"

#include "shaders/intersectionshader.h"
#include "shaders/depthshader.h"
#include "shaders/directshader.h"
#include "shaders/normalshader.h"

#include "shapes/triangle.h"
#include"shapes/sphere.h"

#include <chrono>

int main(int argc, char *argv[])
{
	std::cout << "RTIS - Simple Ray Tracer" << std::endl << std::endl;
	
	std::string fileName;
	if (argc > 3 || argc == 2)
	{
		std::cout << "Usage: RTIS pathToInputFileScene pathOutputFile" << std::endl;
		return 1;
	}
	char *sceneFileName = argv[1];
	char *outputFileName = argv[2];

	Scene scene;
	Shader *shader;
	
	// Load the scene from file
	if (!scene.loadScene(sceneFileName, shader))
		return 0;
	
	// Call the pre-process method of the scene. This is a virtual method which can vary from shader
	//  to shader. It is intended to hold code regarding any pre-processing task which must be performed
	//  so that the actual render method of the shader can be then executed. An example would be creating
	//  and filling an irradiance cache structure.
	auto begin = std::chrono::high_resolution_clock::now();

	shader->preprocess(scene);
	
	// This is the method where the image of the scene is rendered. The way in which the image is created 
	//  depends on the shader chosen.
	shader->render(scene);
	
	// Eventually perform some post-processing on the rendered image. Examples of such an operation would be
	//  (i) to apply a tone mapping operator (TMO) to convert the dynamic range of the rendered image to 
	//  make it more suitable for display in a computer monitor (which are low dynamoc range, LDR). This option is 
	//  currently unavailable as there is no TMO implemented in RTIS. (ii) to apply some denoiser (also not yet
	//  implemented).
	shader->postprocess(scene);

	// Save the image (which is found in scene.cam->film). For now, only bitmap images are supported (thus, we lose
	//  the HDR info since we must compress the dynamic range to 0-255.
	auto end = std::chrono::high_resolution_clock::now();
	auto elapsed = std::chrono::duration_cast<std::chrono::nanoseconds>(end - begin);

	printf("Time measured: %.3f seconds.\n", elapsed.count() * 1e-9);

	scene.cam->film.save(outputFileName);
	    
    return 0;
}
