#include "irradiance.h"
#include <algorithm>
#include "../samplers/hemisphericalsampler.h"

#include <math.h>

irradiance::irradiance() {
};

irradiance::irradiance(Vector3D bgColor, size_t maxDepth, bool show_point, size_t nSamplesArea_, bool indirectOnly_, size_t nSamplesHemis_) :Shader(bgColor, 1) {
	path = PathShader(bgColor, 1, nSamplesArea_, maxDepth, indirectOnly_);
	jump = 20;
	show_points = show_point;
	irradianceCache2 = octree(2, Vector3D(0.0, 0.0, 4.732), Vector3D(-2.0, 2.0, 3.4640), Vector3D(2.0, -2.0, 6.0), 2.0, 2.0, 6.0- 3.4640);
	indirectOnly = indirectOnly_;
	nSamplesHemis = nSamplesHemis_;
	preprocessDone = false;

	useOctree = false;
};

void irradiance::preprocess(Scene& scene) {
	unsigned int sizeBar = 40;

	// Create a vector with the displacement of each pixel sample
	// Compute the number of samples in one dimension
	size_t nSamples1D = (size_t)std::sqrt(nSamplesPP);
	nSamplesPP = nSamples1D * nSamples1D;
	std::cout << "Using " << nSamplesPP << " samples per pixel" << std::endl;
	// Compute the sub-pixel size
	double subPixelSize = 1.0 / nSamples1D;
	// Compute the sample offset in 1D w.r.t. the subpixel beginning
	double dPixel = subPixelSize / 2.0;

	// Get the film resolution
	size_t resX = scene.getCamera().film.getWidth();
	size_t resY = scene.getCamera().film.getHeight();

	std::vector<Vector3D> displacementSamples;
	for (size_t sampleIndex = 0; sampleIndex < nSamplesPP; sampleIndex++)
	{
		double xCoord, yCoord;
		xCoord = dPixel + (sampleIndex % nSamples1D) * subPixelSize;
		yCoord = dPixel + (sampleIndex / nSamples1D) * subPixelSize;
		Vector3D displacement(xCoord / resX, yCoord / resY, 0);
		displacementSamples.push_back(displacement);
	}

	// Main raytracing loop
	// Out-most loop invariant: we have rendered lin lines 
	for (size_t lin = 0; lin < resY; lin++) {
		if (lin % (resY / sizeBar) == 0)
			std::cout << ".";

		// Inner loop invariant: we have rendered col columns 
		for (size_t col = 0; col < resX; col++)
		{
			// Generate image plane sample: (0,0) is the upper left corner
			// Formalize the different spaces used:
			// - Sample space
			// - NDC (normalized device coordinates)
			// - Screen space
			// - Camera space
			Vector3D pixelColor(0.0);

			Vector3D sampleNDC((double)(col) / resX, (double)(lin) / resY, 0);

			Ray cameraRay = scene.getCamera().generateRay(sampleNDC.x, sampleNDC.y);
			if (lin == 0 and col == 0) {
				pixelColor += computeRecord(lin, col, scene.getCamera().film.getWidth(), scene.getCamera().film.getHeight(), scene, displacementSamples);
			}
			else {
				pixelColor += computeColor3D(cameraRay, scene, col, lin, displacementSamples);
			}

		}
	}

	if (useOctree) {
		std::cout << irradianceCache2.num << std::endl;
	}
	else {
		std::cout << irradianceCache.size() << std::endl;
	}

	preprocessDone = true;

}

Vector3D irradiance::computeRecord(size_t x, size_t y, size_t width_, size_t height_, Scene& scene, std::vector<Vector3D> displacementSamples) {
	Vector3D pixelColor(0.0);

	Intersection its;
	double harmonicMean = 0.0;
	Vector3D sampleNDC((double)(x) / width_, (double)(y) / height_, 0);
	Vector3D result;
	Ray r = scene.getCamera().generateRay(sampleNDC.x, sampleNDC.y);
	scene.getClosestIntersection(r, its);
	for (size_t sampleIndex = 0; sampleIndex < 1; sampleIndex++)
	{
		Vector3D sampleNDC((double)(x + 0.5) / width_, (double)(y + 0.5) / height_, 0);
		Vector3D irradiance;
		//sampleNDC += Vector3D(0.5,0.5,0)

		double u = sampleNDC.x;
		double v = sampleNDC.y;
		Ray cameraRay = scene.getCamera().generateRay(u, v);
		// Compute ray color

		if (scene.getClosestIntersection(r, its))
		{
			/* *************************************************** */
			/* *** Treat the case of diffuse/glossy reflection *** */
			/* *************************************************** */
			if (its.shape->getMaterial().hasDiffuseOrGlossy())
			{
				/* SAMPLE THE HEMISHPERE */
				HemisphericalSampler sampler; // This class gives random points on a hemisphere
				for (size_t n = 0; n < nSamplesHemis; ++n) {
					// Generate a random direction on the hemisphere around the normal at the intersection point
					Vector3D rnd_direction = sampler.getSample(its.normal);
					// Create a new ray with this random direction
					Ray indirectRay(its.itsPoint, rnd_direction, r.depth + 1);
					// Compute the incident radiance from that direction
					Vector3D indirectIncidentRadiance = path.computeColor(indirectRay, scene);

					// ATTENTION: ONLY THE INCIDENT LIGHT SHOULD BE STORED. THAT IS WHY THE ABOVE CODE BECOMES:

					irradiance += indirectIncidentRadiance * dot(rnd_direction, its.normal);
					harmonicMean += computeHarmonic(its, indirectRay, scene, x, y);
				}
			}
			// Once all light sources have been taken into account, return the final result
			result = irradiance * (2 * M_PI) / nSamplesHemis;
			harmonicMean = harmonicMean / nSamplesHemis;

			if (its.shape->isEmissiveShape()) {

				if (r.depth == 0) {
					if (!indirectOnly) {
						result += its.shape->getEmission();
					}

				}
			}

		}
		else
		{
			// Found no intersection along the ray
			// Return the background color
			result = bgColor;
		}
	}

	//std::cout << harmonicMean << std::endl;
	// Store the pixel color
	IrradianceCacheRecord record;
	record.its = its;
	record.color = result;
	record.harmonicMean = harmonicMean;
	record.point2D = Vector3D(x, y, 0);
	if (useOctree) {
		irradianceCache2.pushRecord(record);
	}
	else {
		irradianceCache.push_back(record);
	}
	return result;
}

Vector3D irradiance::computeColor3D(const Ray& r, Scene& scene, size_t x, size_t y, std::vector<Vector3D> displacementSamples) {
	int n = 6;
	Intersection its;
	scene.getClosestIntersection(r, its);
	Vector3D E(0, 0, 0);

	std::vector<IrradianceCacheRecord> neighbours;
	if (useOctree) {
		neighbours = near_n_Neighbour_Octree(its, n);
	}
	else {
		neighbours = near_n_Neighbour(its, n);
	}
	
	//std::cout << neighbours.size()<< std::endl;
	Vector3D Esubi(0, 0, 0);
	double wsubi = 0.0;
	double sumBot = 0.0;
	Vector3D sumTop(0, 0, 0);

	Vector3D p = its.itsPoint;

	IrradianceCacheRecord record;
	double dist;

	for (int i = 0; i < neighbours.size();i++) {

		record = neighbours[i];
		//dist = p.norm(record.its.itsPoint);
		dist = (p - record.its.itsPoint).length();

		wsubi = 1.0 / (dist / record.harmonicMean + sqrt(1.0 - its.normal.scalar_product(record.its.normal)));

		/*if (show_points and dist < 0.02) {
			return Vector3D(1, 0, 0);
		}*/

		bool sampleisvalid = sampleisValid(wsubi, neighbours[i], its);
		//std::cout << neighbours[i].harmonicMean << std::endl;
		if (sampleisvalid or preprocessDone) {
			//std::cout << "if" << std::endl;
			sumTop += record.color * wsubi;
			sumBot += wsubi;
		}
		if (dist == 0 and !show_points) {
			E = record.color;

			double u, v;
			its.shape->getTextureCoordinates(its.itsPoint, u, v);
			return Utils::multiplyPerCanal(E, its.shape->getMaterial().getReflectance(its.normal, its.normal, its.normal, u, v));
		}
	}

	if (sumBot > 0 or preprocessDone) {
		E = sumTop / sumBot;
	}
	else {
		E = computeRecord(x, y, scene.getCamera().film.getWidth(), scene.getCamera().film.getHeight(), scene, displacementSamples);
	}
	
	if(preprocessDone){
		double u, v;
		its.shape->getTextureCoordinates(its.itsPoint, u, v);
		return Utils::multiplyPerCanal(E, its.shape->getMaterial().getReflectance(its.normal, its.normal, its.normal, u, v));
	}
	return E;
}

std::vector <IrradianceCacheRecord> irradiance::near_n_Neighbour(Intersection its, int n) {

	std::vector <std::pair<double, int>>  neighboursNpair;
	std::vector <IrradianceCacheRecord> neighboursN;
	double dist;

	for (int i = 0; i < irradianceCache.size(); i++) {
		dist = its.itsPoint.norm(irradianceCache[i].its.itsPoint);
		neighboursNpair.push_back({ dist ,i });
	}
	/*neighboursN = irradianceCache2.getRecordsOnOct(its.itsPoint);
	if (neighboursN.size() != 0) {
		std::cout << neighboursN.size() << std::endl;
	}*/

	if (neighboursNpair.size() < n) {
		n = neighboursNpair.size();
	}

	std::sort(neighboursNpair.begin(), neighboursNpair.end());
	for (int i = 0; i < n; i++) {
		neighboursN.push_back(irradianceCache[neighboursNpair[i].second]);
	}
	return neighboursN;

}

std::vector <IrradianceCacheRecord> irradiance::near_n_Neighbour_Octree(Intersection its, int n) {
	//std::cout << "near octree" << std::endl;
	return irradianceCache2.getRecordsOnOct(its.itsPoint);
}

double irradiance::computeHarmonic(Intersection its, Ray r, Scene& scene, int x, int y) {
	Intersection its2;
	double harmonicMean = 0.0;
	if (scene.hasIntersection(r)) {
		scene.getClosestIntersection(r, its2);
		harmonicMean = its.itsPoint.norm(its2.itsPoint);
	}
	return harmonicMean;
}

void irradiance::render(Scene& scene) {
	unsigned int sizeBar = 40;

	// Create a vector with the displacement of each pixel sample
	// Compute the number of samples in one dimension
	size_t nSamples1D = (size_t)std::sqrt(nSamplesPP);
	nSamplesPP = nSamples1D * nSamples1D;
	std::cout << "Using " << nSamplesPP << " samples per pixel" << std::endl;
	// Compute the sub-pixel size
	double subPixelSize = 1.0 / nSamples1D;
	// Compute the sample offset in 1D w.r.t. the subpixel beginning
	double dPixel = subPixelSize / 2.0;

	// Get the film resolution
	size_t resX = scene.getCamera().film.getWidth();
	size_t resY = scene.getCamera().film.getHeight();

	std::vector<Vector3D> displacementSamples;
	for (size_t sampleIndex = 0; sampleIndex < nSamplesPP; sampleIndex++)
	{
		double xCoord, yCoord;
		xCoord = dPixel + (sampleIndex % nSamples1D) * subPixelSize;
		yCoord = dPixel + (sampleIndex / nSamples1D) * subPixelSize;
		Vector3D displacement(xCoord / resX, yCoord / resY, 0);
		displacementSamples.push_back(displacement);
	}

	// Main raytracing loop
	// Out-most loop invariant: we have rendered lin lines 
	for (size_t lin = 0; lin < resY; lin++) {
		if (lin % (resY / sizeBar) == 0)
			std::cout << ".";

		// Inner loop invariant: we have rendered col columns 
		for (size_t col = 0; col < resX; col++)
		{
			// Generate image plane sample: (0,0) is the upper left corner
			// Formalize the different spaces used:
			// - Sample space
			// - NDC (normalized device coordinates)
			// - Screen space
			// - Camera space
			Vector3D pixelColor(0.0);

			Vector3D sampleNDC((double)(col) / resX, (double)(lin) / resY, 0);

			Ray cameraRay = scene.getCamera().generateRay(sampleNDC.x, sampleNDC.y);
			if (lin == 0 and col == 0) {
				pixelColor += computeRecord(lin, col, scene.getCamera().film.getWidth(), scene.getCamera().film.getHeight(), scene, displacementSamples);
			}
			else {
				pixelColor += computeColor3D(cameraRay, scene, col, lin, displacementSamples);
			}
			
			// Store the pixel color
			if (preprocessDone) {
				scene.getCamera().film.setPixelValue(col, lin, pixelColor);
			}

		}
	}

	if (useOctree) {
		std::cout << irradianceCache2.num << std::endl;
	}
	else {
		std::cout << irradianceCache.size() << std::endl;
	}
}

Vector3D irradiance::computeColor(const Ray& r, Scene& scene) const
{
	return Vector3D();
}

bool irradiance::sampleisValid(double wsubi, IrradianceCacheRecord i, Intersection its) {
	bool dist = its.itsPoint.norm(i.its.itsPoint) < (i.harmonicMean);
	bool weight = wsubi > 1 / 0.06;
	bool behind = (its.itsPoint - i.its.itsPoint).norm(its.normal + i.its.normal) / 2 >= 0;
	//bool corners = !(i.its.itsPoint.x == -2 && i.its.itsPoint.y == -2) and !(i.its.itsPoint.x == 2 && i.its.itsPoint.y == -2) and !(i.its.itsPoint.x == -2 && i.its.itsPoint.y == 2) and !(i.its.itsPoint.x == 2 && i.its.itsPoint.y == 2);
	
	return dist && weight && behind;
}