#include "pathshader.h"
#include "../core/utils.h"
#include "../samplers/hemisphericalsampler.h"
#include <math.h>

PathShader::PathShader()
{ }

PathShader::PathShader(Vector3D bgColor_, size_t nSpp_, size_t nSamplesArea_, size_t maxDepth_, bool indirectOnly_) :
Shader(bgColor_, nSpp_), nSamplesArea(nSamplesArea_), maxDepth(maxDepth_), indirectOnly(indirectOnly_)
{ }

// Auxiliary function for sampling the point light sources
Vector3D samplePointLightSources(Scene scene, Ray r, Intersection its) {
	Vector3D diffuseOrGlossyColor_Direct_PLS(0);
	// Get the direct contribution of each light
	for (size_t lightSourceIndex = 0; lightSourceIndex < scene.lightSourceList->size(); lightSourceIndex++)
	{
		// 0 - Get the current light source
		const LightSource &ls = scene.lightSourceList->at(lightSourceIndex);
		// 1 - Compute the distance to the lightsource and a normalized vector from the
		// intersection point to the light source
		Vector3D lightDirection = ls.getPosition() - its.itsPoint;
		double lightSourceDistance = lightDirection.length();
		lightDirection = lightDirection.normalized();
		// Construct the shaddow ray (ADD EPSILON)
		//   - Note that the ray segment tested ends at lightSourceDistance
		//   - If any intersection is found, the search for occlusion should end
		Ray shadowRay(its.itsPoint, lightDirection, r.depth + 1, Epsilon, lightSourceDistance);
		if (!scene.hasIntersection(shadowRay))
		{
			// 3 - Compute the Radiance arriving from that light source
			// to the intersection point
			Vector3D incidentRadiance = ls.getRadiance(its.itsPoint);
			// 4 - Compute the reflectance at the intersection point
			const Material &mat = its.shape->getMaterial();
			// 4.1 - Get the texture coordinates
			double u, v;
			its.shape->getTextureCoordinates(its.itsPoint, u, v);
			Vector3D reflectance = mat.getReflectance(its.normal,
				-r.d,
				lightDirection, u, v);
			// 5 - Compute the reflected radiance at p in direction wo and
			// accumulate the contribution of the current light source
			diffuseOrGlossyColor_Direct_PLS += Utils::multiplyPerCanal(incidentRadiance, reflectance) * dot(lightDirection, its.normal);
		}
	}
	return diffuseOrGlossyColor_Direct_PLS;
}

// Auxiliary function used to sample the area light sources
Vector3D sampleAreaLightSources(Scene scene, Ray r, Intersection its, size_t nSamplesArea) {
	Vector3D estimateAreaLS(0);
	for (size_t sample_i = 0; sample_i < nSamplesArea; sample_i++) {
		// Randomly select an area light source with uniform probability
		float psi1 = (float)std::rand() / RAND_MAX; // Get an i.i.d. random numbers between 0-1
		size_t indexLS = round(psi1 * (scene.areaLSList->size() - 1));
		Shape *areaLS = scene.areaLSList->at(indexLS);
		float probaLS = 1. / scene.areaLSList->size();
		// Get a random point on the area light source
		Vector3D point_y, normal_y;
		float probaPoint;
		areaLS->getRandomPointOnSurface(point_y, normal_y, probaPoint);
		// Compute the direction towards that point
		float lightSourceDistance = (point_y - its.itsPoint).length() - 2 * Epsilon;
		Vector3D dirPointInLS = (point_y - its.itsPoint).normalized();
		// Set up the secondary ray (shadow ray)
		Ray secondaryRay(its.itsPoint, dirPointInLS, r.depth + 1, Epsilon, lightSourceDistance);
		// Shoot the ray
		if (!scene.hasIntersection(secondaryRay)) {
			// 4 - Compute the reflectance at the intersection point
			const Material &mat = its.shape->getMaterial();
			// 4.1 - Get the texture coordinates
			double u, v;
			its.shape->getTextureCoordinates(its.itsPoint, u, v);
			Vector3D reflectance = mat.getReflectance(its.normal,
				-r.d,
				secondaryRay.d, u, v);
			// 5 - Compute the reflected radiance at p in direction wo and
			// accumulate the contribution of the current light source
			float dot_y = dot(-secondaryRay.d.normalized(), normal_y);
			if (dot_y < 0) {
				dot_y = dot(-secondaryRay.d.normalized(), -normal_y);
			}
			float dot_x = dot(secondaryRay.d.normalized(), its.normal);
			if (dot_x < 0) {
				dot_x = dot(-secondaryRay.d.normalized(), -its.normal);
			}
			float G = dot_x * dot_y / (lightSourceDistance*lightSourceDistance);
			estimateAreaLS += Utils::multiplyPerCanal(areaLS->getEmission(), reflectance) * G / (probaPoint*probaLS);
		}
	}
	if (nSamplesArea > 0)
		estimateAreaLS = estimateAreaLS / (float)nSamplesArea;

	return estimateAreaLS;
}

Vector3D PathShader::computeColor(const Ray &r, Scene &scene) const
{
    // Get the closest Intersection along the ray
    Intersection its; // Auxiliar structure to store information about the intersection, in case there is one
	if (scene.getClosestIntersection(r, its))
    {
        //Vector3D reflectionColor(0), transmissionColor(0), diffuseOrGlossyColor(0);
		Vector3D reflectionColor(0), transmissionColor(0), diffuseOrGlossyColor_Direct(0), diffuseOrGlossyColor_Indirect(0);
        double reflectanceCoeff  = 1;
        double transmissionCoeff = 0;

        // Boolean which indicates if we are inside the transmiting material
        bool inside = false;
        double cosThetaI = dot(its.normal, -r.d);
        if(cosThetaI < 0)
        {
            inside = true;
        }

        /* ************************************** */
        /* *** Treat the case of transmission *** */
        /* ************************************** */
        if(its.shape->getMaterial().hasTransmission()) {

            // Ratio between the the indexes of refraction of the vacuum and
            // the transmiting material
            double eta = 1.0 / its.shape->getMaterial().getIndexOfRefraction();
            Vector3D normal = its.normal;

            // Check wheter we are inside of a transmiting material
            if(inside)
            {
                // Flip the normal
                normal = -normal;
                // Invert the refraction ratio
                eta = 1.0 / eta;
                cosThetaI = dot(normal, -r.d);
            }

            double cosThetaT;

            // Verify if there is transmission and, if there is, get the
            //  transmission ray angle in "cosThetaT"
            if( !Utils::isTotalInternalReflection(eta, cosThetaI, cosThetaT))
            {
                // Compute the transmission direction
                Vector3D transmissionDir = Utils::computeTransmissionDirection(
                            r, normal, eta, cosThetaI, cosThetaT );

                // 3 - Construct the transmission ray
                Ray transmissionRay(its.itsPoint, transmissionDir, r.depth+1);

                // 4 - Compute the color arriving through the refracted ray
                Vector3D fakeVector;
                double fakeU, fakeV;
                fakeU = fakeV = 0;
                Vector3D color = its.shape->getMaterial().getReflectance(fakeVector, fakeVector, fakeVector, fakeU, fakeV);
				transmissionColor = Utils::multiplyPerCanal(color, computeColor(transmissionRay, scene));

                // 5 - Compute the reflectance coefficient
                reflectanceCoeff = Utils::computeReflectanceCoefficient(eta, cosThetaI, cosThetaT);
            }
            // If there is no transmission
            else
            {
                // We are in Total internal reflection
                reflectanceCoeff = 1.0;
            }

            // 5.4 - The transmission coefficient is the dual of the reflectance coefficient
            transmissionCoeff = 1.0 - reflectanceCoeff;

        }

        /* ********************************************* */
        /* *** Treat the case of specular reflection *** */
        /* ********************************************* */
        if(its.shape->getMaterial().hasSpecular() && !inside)
        {
            // Compute the perfect reflection direction
            Vector3D reflectionDir = Utils::computeReflectionDirection(r.d, its.normal);

            // Construct a new ray from the intersections point
            //  in the perfect reflection direction
            Ray reflectionRay(its.itsPoint, reflectionDir, r.depth+1);
            if(its.shape->getMaterial().hasTransmission())
            {
                Vector3D fakeVector;
                double fakeUV = 0;
                Vector3D color = its.shape->getMaterial().getReflectance(fakeVector, fakeVector, fakeVector, fakeUV, fakeUV);
                reflectionColor = Utils::multiplyPerCanal(color, computeColor(reflectionRay, scene));
            }
            else
            {
                reflectionColor = computeColor(reflectionRay, scene);
            }
        }

		/* *************************************************** */
		/* *** Treat the case of diffuse/glossy reflection *** */
		/* *************************************************** */
        if(its.shape->getMaterial().hasDiffuseOrGlossy())
        {
			/* DIRECT ILLUMINATION */
			if (r.depth > 0 || (r.depth == 0 && !indirectOnly)) {

				if (r.depth < maxDepth) {
					/* DIRECT ILLUMINATION (FROM POINT LIGHT SOURCES) */
					diffuseOrGlossyColor_Direct += samplePointLightSources(scene, r, its);

					/* DIRECT ILLUMINATION (FROM AREA LIGHT SOURCES) */
					diffuseOrGlossyColor_Direct += sampleAreaLightSources(scene, r, its, nSamplesArea);
				}
			}
			/* INDIRECT ILLUMINATION (CONTINUE THE PATH) */
			if (r.depth < maxDepth) {
				// Generate a random direction on the hemisphere around the normal at the intersection point
				HemisphericalSampler sampler;
				Vector3D rnd_direction = sampler.getSample(its.normal);
				// Create a new ray with this random direction
				Ray indirectRay(its.itsPoint, rnd_direction, r.depth + 1);
				// Compute the incident radiance from that direction
				Vector3D indirectIncidentRadiance = computeColor(indirectRay, scene);
				const Material &mat = its.shape->getMaterial();
				double u, v;
				its.shape->getTextureCoordinates(its.itsPoint, u, v);
				Vector3D reflectance = mat.getReflectance(its.normal, -r.d, rnd_direction, u, v);
				diffuseOrGlossyColor_Indirect += Utils::multiplyPerCanal(indirectIncidentRadiance, reflectance) * dot(rnd_direction, its.normal)*(2 * M_PI);
			}
        }
        // Once all light sources have been taken into account, return the final result
		Vector3D result = diffuseOrGlossyColor_Indirect + diffuseOrGlossyColor_Direct + reflectionColor * reflectanceCoeff + transmissionColor * transmissionCoeff;
		if (its.shape->isEmissiveShape()){
			
			if (r.depth == 0) {
				if (!indirectOnly)
					result += its.shape->getEmission();
			}
			else if (r.depth == 1) {
				if (!indirectOnly && nSamplesArea == 0)
					result += its.shape->getEmission();
			}
			else { // r.depth > 1
				if(nSamplesArea == 0)
					result += its.shape->getEmission();
			}
		}
		return result;

    }
    else
    {
        // Found no intersection along the ray
        // Return the background color
        return bgColor;
    }
}
