#include "triangle.h"

Triangle::Triangle(Vector3D p0_, Vector3D p1_, Vector3D p2_, Material* material_)
	: Shape(Matrix4x4(), material_), p0(p0_), p1(p1_), p2(p2_)
{
	normal = cross(p2 - p0, p1 - p0).normalized();
}

void Triangle::getRandomPointOnSurface(Vector3D &point_out, Vector3D &normal_out, float &proba) const{
	float r1 = (float)std::rand() / RAND_MAX;
	float r2 = (float)std::rand() / RAND_MAX;
	float aux = sqrt(r1);
	point_out =  p0*(1 - aux) + p1*(aux*(1-r2)) + p2*(r2*aux);
	normal_out = normal;
	float area = cross(p1-p0, p2-p0).length()/2.;
	proba = 1. / area;
}

bool Triangle::rayIntersect(const Ray & ray, Intersection & its) const
{
	 // Compute the denominator of the tHit formula
	double denominator = dot(ray.d, normal);
	
    // If the denominator is very close to zero, then the ray and the
    // plane are almost parallel
    if(std::abs(denominator) < Epsilon)
        return false;

    // Effectivelly compute the intersection point
	double tHit = dot((p0 - ray.o), normal) / denominator;

    // Is tHit outside the bounds of the ray segment we want to test intersecion with?
    if (tHit < ray.minT || tHit > ray.maxT)
        return false;

	//Test if intersection point is inside the triangle
	Vector3D itsPoint = ray.o + ray.d * tHit;

	//Test p1 - p0
	if (dot(cross(itsPoint - p0, p1 - p0), normal) < 0)
		return false;

	//Test p2 - p1
	if (dot(cross(itsPoint - p1, p2 - p1), normal) < 0)
		return false;

	//Test p0 - p2
	if (dot(cross(itsPoint - p2, p0 - p2), normal) < 0)
		return false;

	//Store intersection info
	its.itsPoint = itsPoint;
	if (dot(-ray.d, normal) < 0)
		its.normal = -normal;
	else
		its.normal = normal;
	its.shape = this;

    // Arriving here, we know that the ray segment intersects the plan
	ray.maxT = tHit;
    return true;
}

bool Triangle::shadowRayIntersect(const Ray& ray) const
{
	// Compute the denominator of the tHit formula
	double denominator = dot(ray.d, normal);

	// If the denominator is very close to zero, then the ray and the
	// plane are almost parallel
	if (std::abs(denominator) < Epsilon)
		return false;

	// Effectivelly compute the intersection point
	double tHit = dot((p0 - ray.o), normal) / denominator;

	// Is tHit outside the bounds of the ray segment we want to test intersecion with?
	if (tHit < ray.minT || tHit > ray.maxT)
		return false;

	//Test if intersection point is inside the triangle
	Vector3D itsPoint = ray.o + ray.d * tHit;

	//Test p1 - p0
	if (dot(cross(itsPoint - p0, p1 - p0), normal) < 0)
		return false;

	//Test p2 - p1
	if (dot(cross(itsPoint - p1, p2 - p1), normal) < 0)
		return false;

	//Test p0 - p2
	if (dot(cross(itsPoint - p2, p0 - p2), normal) < 0)
		return false;

	// Arriving here, we know that the ray segment intersects the plan
	return true;
}
